/* BBLoSlitta - a BB4W plugin to put other plugins in the slit.
   Copyright (C) 2004 kana <nicht AT s8 DOT xrea DOT con>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>


#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))




/* Constants */

static const char BBLS_NAME[] = "BBLoSlitta";
static const char BBLS_VERSION[] = "0.0.0";
static const char BBLS_AUTHOR[] = "kana";
static const char BBLS_RELEASE[] = "2004-10-22";
static const char BBLS_LINK[] = "http://nicht.s8.xrea.com/2004/10/BBLoSlitta";
static const char BBLS_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static const char BBLS_CLASS_NAME[] = "BBLoSlitta";
static const char BBLS_RC_NAME[] = "BBLoSlitta.rc";




/* Global variables */

	/**
	 * These are set in beginPlugin().
	 * And these will never be changed until endPlugin().
	 */
static HWND G_PluginWnd = NULL;
static HMODULE G_PluginModule = NULL;
static HWND G_SlitWnd = NULL;
static HHOOK G_Hook = NULL;


typedef struct {
	char* module_name;
	char* class_name;
} Pattern;

static Pattern* G_PatternTable = NULL;
int G_PatternTableNum = 0;

static HWND* G_TargetTable = NULL;
int G_TargetTableNum = 0;








/* Misc. functions */

static void
VSNPRINTF(char* buf, int size, const char* format, va_list va)
{
	if (buf == NULL || size <= 0) return;

	vsnprintf(buf, size, format, va);

	buf[size - 1] = '\0';
}




static void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	VSNPRINTF(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( G_PluginWnd, buf, BBLS_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TOPMOST);
}




static BOOL
is_2k_or_later(void)
{
	OSVERSIONINFO os;

	os.dwOSVersionInfoSize = sizeof(os);
	GetVersionEx(&os);

	return 5 <= os.dwMajorVersion;
}




static BOOL
get_module_dir(char* buf, int size)
{
	char* bs;

	if (!GetModuleFileName(G_PluginModule, buf, size))
		return FALSE;
	buf[size - 1] = '\0';

	bs = strrchr(buf, '\\');
	if (bs != NULL)
		bs[1] = '\0';

	return TRUE;
}


static BOOL
get_rc_path(char* buf, int size, const char* rc_name)
{
	char dir[MAX_PATH];

	if (!get_module_dir(dir, NUMBER_OF(dir)))
		return FALSE;

	if (size < strlen(dir) + strlen(rc_name) + 1)
		return FALSE;

	strcpy(buf, dir);
	strcat(buf, rc_name);

	return TRUE;
}




static BOOL
is_target_window(HWND hwnd)
{
	int i;
	char buf[256];
	HMODULE hmodule;
	char* p;

	for (i = 0; i < G_PatternTableNum; i++) {
		if (G_PatternTable[i].class_name[0] != '\0') {
			if (!GetClassName(hwnd, buf, NUMBER_OF(buf)))
				continue;

			if (0 != strcmp(buf, G_PatternTable[i].class_name))
				continue;
		}

		if (G_PatternTable[i].module_name[0] != '\0') {
			hmodule = (HMODULE)GetClassLong(hwnd, GCL_HMODULE);
			if (hmodule == NULL)
				continue;

			if (!GetModuleFileName(hmodule, buf, NUMBER_OF(buf)))
				continue;

			p = strrchr(buf, '\\');
			if (p != NULL) {
				memmove( buf, p + 1,
				         sizeof(char) * (strlen(p+1) + 1) );
			}

			if (0 != strcmp(buf, G_PatternTable[i].module_name))
				continue;
		}

		return TRUE;
	}

	return FALSE;
}


static BOOL
is_slitted_window(HWND hwnd)
{
	int i;

	for (i = 0; i < G_TargetTableNum; i++) {
		if (hwnd == G_TargetTable[i])
			return TRUE;
	}

	return FALSE;
}




static void
PatternTable_Del(Pattern* table)
{
	int i;

	for (i = 0; i < G_PatternTableNum; i++) {
		free(G_PatternTable[i].module_name);
		free(G_PatternTable[i].class_name);
	}
	free(G_PatternTable);
}




static void
TargetTable_Del(hwnd)
{
	int i;

	for (i = 0; i < G_TargetTableNum; i++) {
		PostMessage( G_SlitWnd, SLIT_REMOVE, 0,
		             (LPARAM)G_TargetTable[i] );
	}

	free(G_TargetTable);
}


static void
TargetTable_Add(HWND hwnd)
{
	HWND* t;

	t = realloc( G_TargetTable,
	             sizeof(G_TargetTable[0]) * (G_TargetTableNum + 1) );
	if (t == NULL) {
		notice("%s: Out of memory", __func__);
		return;
	}

	G_TargetTable = t;
	G_TargetTable[G_TargetTableNum] = hwnd;
	G_TargetTableNum++;

	PostMessage(G_SlitWnd, SLIT_ADD, 0, (LPARAM)hwnd);
}


static void
TargetTable_Remove(HWND hwnd)
{
	int i;

	for (i = 0; i < G_TargetTableNum; i++) {
		if (hwnd == G_TargetTable[i]) {
			PostMessage(G_SlitWnd, SLIT_REMOVE, 0, (LPARAM)hwnd);

			G_TargetTable[i] = G_TargetTable[G_TargetTableNum - 1];
			G_TargetTableNum--;
		}
	}
}


static BOOL CALLBACK
_TargetTable_AddExistentWindows_EnumProc(HWND hwnd, LPARAM lp)
{
	if (is_target_window(hwnd))
		PostMessage(G_PluginWnd, SLIT_ADD, 0, (LPARAM)hwnd);

	return TRUE;
}

static void
TargetTable_AddExistentWindows(void)
{
	EnumWindows(_TargetTable_AddExistentWindows_EnumProc, 0);
}




static void
read_settings(void)
{
	char rc[MAX_PATH];
	FILE* file;
	char line[1024];
	char* p;
	int lineno;

	if (!get_rc_path(rc, NUMBER_OF(rc), BBLS_RC_NAME)) {
		notice("Failed to get the path to %s", BBLS_RC_NAME);
		goto E_RC_PATH;
	}

	file = fopen(rc, "rt");
	if (file == NULL) {
		notice("Failed to open %s", BBLS_RC_NAME);
		goto E_FOPEN;
	}

	lineno = 0;
	while ((p = fgets(line, NUMBER_OF(line), file)) != NULL) {
		char* delim;
		char* module_name;
		char* class_name;
		Pattern* t;

		lineno++;

		while (*p == ' ' || *p == '\t')
			p++;

		if (*p == '#' || *p == '\n')
			continue;

		delim = strchr(p, ':');
		if (delim == NULL) {
			notice("%s:%d: Syntax error", BBLS_RC_NAME, lineno);
			continue;
		}
		*delim = '\0';

		delim++;
		delim[strlen(delim) - 1] = '\0';

		module_name = strdup(p);
		class_name = strdup(delim);
		if (module_name == NULL || class_name == NULL) {
			notice("%s:%d: Out of memory", BBLS_RC_NAME, lineno);
			free(module_name);
			free(class_name);
			break;
		}


		t = realloc( G_PatternTable,
		             sizeof(G_PatternTable[0])*(G_PatternTableNum+1) );
		if (t == NULL) {
			notice("%s:%d: Out of memory", BBLS_RC_NAME, lineno);
			free(module_name);
			free(class_name);
			break;
		}

		t[G_PatternTableNum].module_name = module_name;
		t[G_PatternTableNum].class_name = class_name;
		G_PatternTable = t;
		G_PatternTableNum++;
	}

E_FOPEN:
E_RC_PATH:
	return;
}








/* Main functions */

static LRESULT CALLBACK
BBLS_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case SLIT_ADD:
		TargetTable_Add((HWND)lp);
		break;

	case SLIT_REMOVE:
		TargetTable_Remove((HWND)lp);
		break;

	case SLIT_UPDATE:
		PostMessage(G_SlitWnd, SLIT_UPDATE, 0, 0);
		break;


	case WM_CREATE:
		break;

	case WM_DESTROY:
		break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}




LRESULT CALLBACK
BBLS_CBTProc(int ncode, WPARAM wp, LPARAM lp)
{
	if (ncode == HCBT_CREATEWND) {
		if (is_target_window((HWND)wp))
			PostMessage(G_PluginWnd, SLIT_ADD, 0, (LPARAM)wp);
	} else if (ncode == HCBT_DESTROYWND) {
		if (is_slitted_window((HWND)wp))
			PostMessage(G_PluginWnd, SLIT_REMOVE, 0, (LPARAM)wp);
	} else if (ncode == HCBT_MOVESIZE) {
		if (is_slitted_window((HWND)wp))
			PostMessage(G_PluginWnd, SLIT_UPDATE, 0, 0);
	} else {
		; /* nop */
	}

	return CallNextHookEx(NULL, ncode, wp, lp);
}








/* Public functions */

DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	G_PluginModule = hplugininstance;

	if (G_SlitWnd == NULL) {
		notice("The slit is not found");
		goto E_SLIT_WND;
	}

	read_settings();

	/* Create the window of this plugin */
	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = BBLS_WndProc;
	wc.hInstance = hplugininstance;
	wc.lpszClassName = BBLS_CLASS_NAME;
	if (!RegisterClass(&wc)) {
		notice("Failed RegisterClass()");
		goto E_REGISTER_CLASS;
	}

	G_PluginWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                              BBLS_CLASS_NAME,
	                              NULL,
	                              0,
	                              0, 0, 0, 0,
	                              (is_2k_or_later() ? HWND_MESSAGE : NULL),
	                              NULL,
	                              hplugininstance,
	                              NULL );
	if (G_PluginWnd == NULL) {
		notice("Failed CreateWindowEx()");
		goto E_CREATE_WINDOW;
	}

	G_Hook = SetWindowsHookEx( WH_CBT, BBLS_CBTProc, NULL,
	                           GetCurrentThreadId() );
	if (G_Hook == NULL) {
		notice("Failed SetWindowsHookEx()");
		goto E_HOOK;
	}

	TargetTable_AddExistentWindows();

	return 0;


E_HOOK:
	DestroyWindow(G_PluginWnd);
E_CREATE_WINDOW:
	UnregisterClass(BBLS_CLASS_NAME, hplugininstance);
E_REGISTER_CLASS:
	PatternTable_Del(G_PatternTable);
E_SLIT_WND:
	return !0;
}


DLL_EXPORT int
beginSlitPlugin(HINSTANCE hplugininstance, HWND slitwindow)
{
	G_SlitWnd = slitwindow;
	return beginPlugin(hplugininstance);
}


DLL_EXPORT int
beginPluginEx(HINSTANCE hplugininstance, HWND slitwindow)
{
	return beginSlitPlugin(hplugininstance, slitwindow);
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	UnhookWindowsHookEx(G_Hook);

	DestroyWindow(G_PluginWnd);
	UnregisterClass(BBLS_CLASS_NAME, hplugininstance);

	TargetTable_Del(G_TargetTable);
	PatternTable_Del(G_PatternTable);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	default: /* FALLTHRU */
	case PLUGIN_NAME: return BBLS_NAME;
	case PLUGIN_VERSION: return BBLS_VERSION;
	case PLUGIN_AUTHOR: return BBLS_AUTHOR;
	case PLUGIN_RELEASE: return BBLS_RELEASE;
	case PLUGIN_LINK: return BBLS_LINK;
	case PLUGIN_EMAIL: return BBLS_EMAIL;
	}
}




/* __END__ */
